#lang racket/base

(require (planet tonyg/bitsyntax))
(require (planet vyzo/crypto:2:3))

(require racket/set)
(require racket/match)

(require "oakley-groups.rkt")

(require "ssh-host-key.rkt")

(require "ssh-numbers.rkt")
(require "ssh-message-types.rkt")
(require "ssh-exceptions.rkt")
(require "ssh-transport.rkt")

(require "os2-support.rkt")

(provide rekey-interval
	 rekey-volume
	 ssh-session)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Data definitions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; A RekeyState is one of
;; - a (rekey-wait Number Number), representing a time or
;;   transfer-amount by which rekeying should be started
;; - a (rekey-local SshMsgKexinit), when we've sent our local
;;   algorithm list and are waiting for the other party to send theirs
;; - a (rekey-in-progress KeyExchangeState), when both our local
;;   algorithm list has been sent and the remote one has arrived and the
;;   actual key exchange has begun
(struct rekey-wait (deadline threshold-bytes) #:transparent)
(struct rekey-local (local-algorithms) #:transparent)
(struct rekey-in-progress (state) #:transparent)

;; An AuthenticationState is one of
;; - #f, for not-yet-authenticated
;; - an (authenticated String String), recording successful completion
;;   of the authentication protocol after a request to be identified
;;   as the given username for the given service.
;; TODO: When authentication is properly implemented, we will need
;; intermediate states here too.
(struct authenticated (username service) #:transparent)

;; A PacketDispatcher is a Hashtable mapping Byte to PacketHandler.

;; A PacketHandler is a (Bytes DecodedPacket ConnectionState -> Transition).
;; The raw received bytes of the packet are given because sometimes
;; cryptographic operations on the received bytes are mandated by the
;; protocol.

;; A ConnectionState is a (connection ... TODO fix this) representing
;; the complete state of the SSH transport, authentication, and
;; connection layers.
(struct connection (discard-next-packet?
		    dispatch-table
		    total-transferred
		    rekey-state
		    authentication-state
		    channel-map
		    is-server?
		    local-id
		    remote-id
		    session-id) ;; starts off #f until initial keying
	#:transparent)

;; A CloseState is one of
;; - 'neither, indicating that neither side has signalled closure
;; - 'local, only the local end has signalled closure
;; - 'remote, only the remote end has signalled closure
;; - 'both, both ends have signalled closure.
;; Represents local knowledge of the state of a shared shutdown state
;; machine.
;;
;;                'neither
;;                /       \
;;              \/         \/
;;           'local      'remote
;;               \         /
;;                \/     \/
;;                  'both

;; A ChannelState is a (ssh-channel ...) TODO
;; Named ssh-channel to avoid conflicts with Racket's built-in
;; synchronous channels.
(struct ssh-channel (my-ref ;; Uint32
		     your-ref ;; Maybe<Uint32>
		     type ;; String
		     continuations ;; TransactionManager (see ordered-rpc.rkt)
		     outbound-window ;; Maybe<Natural>
		     outbound-packet-size ;; Maybe<Natural>
		     inbound-window ;; Natural
		     eof-state ;; CloseState covering EOF signals
		     close-state ;; CloseState covering CLOSE signals
		     )
	#:transparent)

;; Generic inputs into the exchange-hash part of key
;; exchange. Diffie-Hellman uses these fields along with the host key,
;; the exchange values, and the shared secret to get the final hash.
(struct exchange-hash-info (client-id
			    server-id
			    client-kexinit-bytes
			    server-kexinit-bytes)
	#:transparent)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Parameters
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define rekey-interval (make-parameter 3600))
(define rekey-volume (make-parameter 1000000000))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Packet dispatch and handling
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Bytes -> Byte
;; Retrieves the packet type byte from a packet.
(define (encoded-packet-msg-type encoded-packet)
  (bytes-ref encoded-packet 0))

;; PacketDispatcher [ Byte Maybe<PacketHandler> ]* -> PacketDispatcher
;; Adds or removes handlers to or from the given PacketDispatcher.
(define (extend-packet-dispatcher core-dispatcher . key-value-pairs)
  (let loop ((d core-dispatcher)
	     (key-value-pairs key-value-pairs))
    (cond
     ((null? key-value-pairs)
      d)
     ((null? (cdr key-value-pairs))
      (error 'extend-packet-dispatcher
	     "Must call extend-packet-dispatcher with matched key/value pairs"))
     (else
      (loop (let ((packet-type-number (car key-value-pairs))
		  (packet-handler-or-false (cadr key-value-pairs)))
	      (if packet-handler-or-false
		  (hash-set d packet-type-number packet-handler-or-false)
		  (hash-remove d packet-type-number)))
	    (cddr key-value-pairs))))))

;; Transition [ Byte Maybe<PacketHandler> ]* -> ConnectionState
;; Installs (or removes) PacketHandlers in the given connection state;
;; see extend-packet-dispatcher.
(define (set-handlers t . key-value-pairs)
  (extend-transition* t
    (lambda (conn)
      (struct-copy connection conn
		   [dispatch-table (apply extend-packet-dispatcher
					  (connection-dispatch-table conn)
					  key-value-pairs)]))))

;; Transition Byte PacketHandler -> ConnectionState
;; Installs a PacketHandler that removes the installed dispatch entry
;; and then delegates to its argument.
(define (oneshot-handler t packet-type-number packet-handler)
  (set-handlers t
		packet-type-number
		(lambda (packet message conn)
		  (packet-handler packet
				  message
				  (set-handlers conn packet-type-number #f)))))

(define (dispatch-packet seq packet message conn)
  (define packet-type-number (encoded-packet-msg-type packet))
  (if (and (not (rekey-wait? (connection-rekey-state conn)))
	   (or (not (ssh-msg-type-transport-layer? packet-type-number))
	       (= packet-type-number SSH_MSG_SERVICE_REQUEST)
	       (= packet-type-number SSH_MSG_SERVICE_ACCEPT)))
      ;; We're in the middle of some phase of an active key-exchange,
      ;; and received a packet that's for a higher layer than the
      ;; transport layer, or one of the forbidden types given at the
      ;; send of RFC4253 section 7.1.
      (disconnect-with-error SSH_DISCONNECT_PROTOCOL_ERROR
			     "Packets of type ~v forbidden while in key-exchange"
			     packet-type-number)
      ;; We're either idling, or it's a permitted packet type while
      ;; performing key exchange. Look it up in the dispatch table.
      (let ((handler (hash-ref (connection-dispatch-table conn)
			       packet-type-number
			       #f)))
	(if handler
	    (handler packet message conn)
	    (transition conn
	      (send-message (outbound-packet (ssh-msg-unimplemented seq))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Handlers for core transport packet types
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; PacketHandler for handling SSH_MSG_DISCONNECT.
(define (handle-msg-disconnect packet message conn)
  (disconnect-with-error* #t
			  '()
			  (ssh-msg-disconnect-reason-code message)
			  "Received SSH_MSG_DISCONNECT with reason code ~a and message ~s"
			  (ssh-msg-disconnect-reason-code message)
			  (bytes->string/utf-8 (bit-string->bytes
						(ssh-msg-disconnect-description message)))))

;; PacketHandler for handling SSH_MSG_IGNORE.
(define (handle-msg-ignore packet message conn)
  conn)

;; PacketHandler for handling SSH_MSG_UNIMPLEMENTED.
(define (handle-msg-unimplemented packet message conn)
  (disconnect-with-error/local-info
   `((offending-sequence-number ,(ssh-msg-unimplemented-sequence-number message)))
   SSH_DISCONNECT_PROTOCOL_ERROR
   "Disconnecting because of received SSH_MSG_UNIMPLEMENTED."))

;; PacketHandler for handling SSH_MSG_DEBUG.
(define (handle-msg-debug packet message conn)
  (log-debug (format "Received SSHv2 SSH_MSG_DEBUG packet ~v" message))
  conn)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Key Exchange
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (rekey-in-seconds-or-bytes delta-seconds delta-bytes total-transferred)
	;; (transition conn
	;;   (send-message (set-timer 'rekey-timer
	;; 			   (* (rekey-wait-deadline rekey) 1000)
	;; 			   'absolute)))]
  (rekey-wait (+ (current-seconds) delta-seconds)
	      (+ total-transferred delta-bytes)))

(define (time-to-rekey? rekey conn)
  (and (rekey-wait? rekey)
       (or (>= (current-seconds) (rekey-wait-deadline rekey))
	   (>= (connection-total-transferred conn) (rekey-wait-threshold-bytes rekey)))))

;; (SshMsgKexinit -> Symbol) SshMsgKexinit SshMsgKexinit -> Symbol
;; Computes the name of the "best" algorithm choice at the given
;; getter, using the rules from the RFC and the client and server
;; algorithm precedence lists.
(define (best-result getter client-algs server-algs)
  (define client-list0 (getter client-algs))
  (define server-list (getter server-algs))
  (let loop ((client-list client-list0))
    (cond
     ((null? client-list) (disconnect-with-error/local-info
			   `((client-list ,client-list0)
			     (server-list ,server-list))
			   SSH_DISCONNECT_KEY_EXCHANGE_FAILED
			   "Could not agree on a suitable algorithm for ~v"
			   getter))
     ((memq (car client-list) server-list) (car client-list))
     (else (loop (cdr client-list))))))

;; ExchangeHashInfo Bytes Natural Natural Natural -> Bytes
;; Computes the session ID as defined by SSH's DH key exchange method.
(define (dh-exchange-hash hash-info host-key e f k)
  (let ((block-to-hash
	 (bit-string->bytes
	  (bit-string ((exchange-hash-info-client-id hash-info) :: (t:string))
		      ((exchange-hash-info-server-id hash-info) :: (t:string))
		      ((exchange-hash-info-client-kexinit-bytes hash-info) :: (t:string))
		      ((exchange-hash-info-server-kexinit-bytes hash-info) :: (t:string))
		      (host-key :: (t:string))
		      (e :: (t:mpint))
		      (f :: (t:mpint))
		      (k :: (t:mpint))))))
    (sha1 block-to-hash)))

;; ExchangeHashInfo Symbol Symbol ConnectionState
;;   (Bytes Bytes Symbol ConnectionState -> ConnectionState)
;;   -> Transition
;; Performs the server's half of the Diffie-Hellman key exchange protocol.
(define (perform-server-key-exchange hash-info kex-alg host-key-alg conn finish)
  (case kex-alg
    [(diffie-hellman-group14-sha1 diffie-hellman-group1-sha1)
     (define group (if (eq? kex-alg 'diffie-hellman-group14-sha1)
		       dh:oakley-group-14
		       dh:oakley-group-2)) ;; yes, SSH's group1 == Oakley/RFC2409 group 2
     (define-values (private-key public-key) (generate-key group))
     (define public-key-as-integer (bit-string->integer public-key #t #f))
     (oneshot-handler conn
		      SSH_MSG_KEXDH_INIT
		      (lambda (packet message conn)
			(define e (ssh-msg-kexdh-init-e message))
			(define e-width (mpint-width e))
			(define e-as-bytes (integer->bit-string e (* 8 e-width) #t))
			(define shared-secret (compute-key private-key e-as-bytes))
			(define hash-alg sha1)
			(define-values (host-key-private host-key-public)
			  (host-key-algorithm->keys host-key-alg))
			(define host-key-bytes
			  (pieces->ssh-host-key (public-key->pieces host-key-public)))
			(define exchange-hash
			  (dh-exchange-hash hash-info
					    host-key-bytes
					    e
					    public-key-as-integer
					    (bit-string->integer shared-secret #t #f)))
			(define h-signature (host-key-signature host-key-private
								host-key-alg
								exchange-hash))
			(prefix-transition (finish shared-secret exchange-hash hash-alg conn)
			  (send-message (outbound-packet
					 (ssh-msg-kexdh-reply (bit-string->bytes host-key-bytes)
							      public-key-as-integer
							      (bit-string->bytes h-signature)))))))]
    [else (disconnect-with-error SSH_DISCONNECT_KEY_EXCHANGE_FAILED
				 "Bad key-exchange algorithm ~v" kex-alg)]))

;; ExchangeHashInfo Symbol Symbol ConnectionState
;;   (Bytes Bytes Symbol ConnectionState -> ConnectionState)
;;   -> Transition
;; Performs the client's half of the Diffie-Hellman key exchange protocol.
(define (perform-client-key-exchange hash-info kex-alg host-key-alg conn finish)
  (case kex-alg
    [(diffie-hellman-group14-sha1 diffie-hellman-group1-sha1)
     (define group (if (eq? kex-alg 'diffie-hellman-group14-sha1)
		       dh:oakley-group-14
		       dh:oakley-group-2)) ;; yes, SSH's group1 == Oakley/RFC2409 group 2
     (define-values (private-key public-key) (generate-key group))
     (define public-key-as-integer (bit-string->integer public-key #t #f))
     (prefix-transition
	 (oneshot-handler conn
			  SSH_MSG_KEXDH_REPLY
			  (lambda (packet message conn)
			    (define f (ssh-msg-kexdh-reply-f message))
			    (define f-width (mpint-width f))
			    (define f-as-bytes (integer->bit-string f (* 8 f-width) #t))
			    (define shared-secret (compute-key private-key f-as-bytes))
			    (define hash-alg sha1)
			    (define host-key-bytes (ssh-msg-kexdh-reply-host-key message))
			    (define host-public-key
			      (pieces->public-key (ssh-host-key->pieces host-key-bytes)))
			    (define exchange-hash
			      (dh-exchange-hash hash-info
						host-key-bytes
						public-key-as-integer
						f
						(bit-string->integer shared-secret #t #f)))
			    (verify-host-key-signature! host-public-key
							host-key-alg
							exchange-hash
							(ssh-msg-kexdh-reply-h-signature message))
			    (finish shared-secret exchange-hash hash-alg conn)))
       (send-message (outbound-packet (ssh-msg-kexdh-init public-key-as-integer))))]
    [else (disconnect-with-error SSH_DISCONNECT_KEY_EXCHANGE_FAILED
				 "Bad key-exchange algorithm ~v" kex-alg)]))

;; PacketHandler for handling SSH_MSG_KEXINIT.
(define (handle-msg-kexinit packet message conn)
  (define rekey (connection-rekey-state conn))
  (when (rekey-in-progress? rekey)
    (disconnect-with-error SSH_DISCONNECT_PROTOCOL_ERROR
			   "Received SSH_MSG_KEXINIT during ongoing key exchange"))
  (define local-algs (if (rekey-local? rekey)
			 (rekey-local-local-algorithms rekey)
			 ((local-algorithm-list))))
  (define encoded-local-algs (ssh-message-encode local-algs))
  (define remote-algs message)
  (define encoded-remote-algs packet)

  (define is-server? (connection-is-server? conn))
  (define c (if is-server? remote-algs local-algs))
  (define s (if is-server? local-algs remote-algs))

  (define kex-alg (best-result ssh-msg-kexinit-kex_algorithms c s))
  (define host-key-alg (best-result ssh-msg-kexinit-server_host_key_algorithms c s))
  (define c2s-enc (best-result ssh-msg-kexinit-encryption_algorithms_client_to_server c s))
  (define s2c-enc (best-result ssh-msg-kexinit-encryption_algorithms_server_to_client c s))
  (define c2s-mac (best-result ssh-msg-kexinit-mac_algorithms_client_to_server c s))
  (define s2c-mac (best-result ssh-msg-kexinit-mac_algorithms_server_to_client c s))
  (define c2s-zip (best-result ssh-msg-kexinit-compression_algorithms_client_to_server c s))
  (define s2c-zip (best-result ssh-msg-kexinit-compression_algorithms_server_to_client c s))
  ;; Ignore languages.
  ;; Don't check the reserved field here, either. TODO: should we?

  (define (guess-matches? chosen-value getter)
    (let ((remote-choices (getter remote-algs)))
      (and (pair? remote-choices) ;; not strictly necessary because of
				  ;; the error behaviour of
				  ;; best-result.
	   (eq? (car remote-choices) ;; the remote peer's guess for this parameter
		chosen-value))))

  (define should-discard-first-kex-packet
    (and (ssh-msg-kexinit-first_kex_packet_follows remote-algs)
	 ;; They've already transmitted their guess. Does their guess match
	 ;; what we've actually selected?
	 (not (and
	       (guess-matches? kex-alg ssh-msg-kexinit-kex_algorithms)
	       (guess-matches? host-key-alg ssh-msg-kexinit-server_host_key_algorithms)
	       (guess-matches? c2s-enc ssh-msg-kexinit-encryption_algorithms_client_to_server)
	       (guess-matches? s2c-enc ssh-msg-kexinit-encryption_algorithms_server_to_client)
	       (guess-matches? c2s-mac ssh-msg-kexinit-mac_algorithms_client_to_server)
	       (guess-matches? s2c-mac ssh-msg-kexinit-mac_algorithms_server_to_client)
	       (guess-matches? c2s-zip ssh-msg-kexinit-compression_algorithms_client_to_server)
	       (guess-matches? s2c-zip ssh-msg-kexinit-compression_algorithms_server_to_client)))))

  (define (continue-after-discard conn)
    ((if is-server?
	 perform-server-key-exchange
	 perform-client-key-exchange)
     (if is-server?
	 (exchange-hash-info (connection-remote-id conn)
			     (connection-local-id conn)
			     encoded-remote-algs
			     encoded-local-algs)
	 (exchange-hash-info (connection-local-id conn)
			     (connection-remote-id conn)
			     encoded-local-algs
			     encoded-remote-algs))
     kex-alg
     host-key-alg
     conn
     continue-after-key-exchange))

  (define (continue-after-key-exchange shared-secret exchange-hash hash-alg conn)
    (define session-id (if (connection-session-id conn)
			   (connection-session-id conn) ;; don't overwrite existing ID
			   exchange-hash))
    (define k-h-prefix (bit-string ((bit-string->integer shared-secret #t #f) :: (t:mpint))
				   (exchange-hash :: binary)))
    (define (derive-key kind needed-bytes-or-false)
      (let extend ((key (hash-alg (bit-string->bytes
				   (bit-string (k-h-prefix :: binary)
					       (kind :: binary)
					       (session-id :: binary))))))
	(cond
	 ((eq? #f needed-bytes-or-false)
	  key)
	 ((>= (bytes-length key) needed-bytes-or-false)
	  (subbytes key 0 needed-bytes-or-false))
	 (else
	  (extend (bytes-append key (hash-alg (bit-string->bytes
					       (bit-string (k-h-prefix :: binary)
							   (key :: binary))))))))))
    (oneshot-handler (struct-copy connection conn
				  [session-id session-id]) ;; just in case it changed
		     SSH_MSG_NEWKEYS
		     (lambda (newkeys-packet newkeys-message conn)
		       ;; First, send our SSH_MSG_NEWKEYS,
		       ;; incrementing the various counters, and then
		       ;; apply the new algorithms.
		       (transition
			   (set-handlers (struct-copy connection conn
						      [rekey-state
						       (rekey-in-seconds-or-bytes
							(rekey-interval)
							(rekey-volume)
							(connection-total-transferred conn))])
					 SSH_MSG_SERVICE_REQUEST handle-msg-service-request)
			 (send-message (outbound-packet (ssh-msg-newkeys)))
			 (send-message
			  (new-keys (connection-is-server? conn)
				    derive-key
				    c2s-enc s2c-enc
				    c2s-mac s2c-mac
				    c2s-zip s2c-zip))))))

  (let ((t (if should-discard-first-kex-packet
	       (struct-copy connection (continue-after-discard conn) [discard-next-packet? #t])
	       (continue-after-discard conn))))
    (prefix-transition* t
      (lambda (conn)
	(if (rekey-wait? (connection-rekey-state conn))
	    (transition (struct-copy connection conn [rekey-state (rekey-local local-algs)])
	      (send-message (outbound-packet local-algs)))
	    conn)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Service request manager
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (handle-msg-service-request packet message conn)
  (define service (bit-string->bytes (ssh-msg-service-request-service-name message)))
  (match service
    [#"ssh-userauth"
     (if (connection-authentication-state conn)
	 (disconnect-with-error SSH_DISCONNECT_SERVICE_NOT_AVAILABLE
				"Repeated authentication is not permitted")
	 (prefix-transition (oneshot-handler conn
					     SSH_MSG_USERAUTH_REQUEST
					     handle-msg-userauth-request)
	   (send-message (outbound-packet (ssh-msg-service-accept service)))))]
    [else
     (disconnect-with-error SSH_DISCONNECT_SERVICE_NOT_AVAILABLE
			    "Service ~v not supported"
			    service)]))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; User authentication
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (handle-msg-userauth-request packet message conn)
  (define user-name (bit-string->bytes (ssh-msg-userauth-request-user-name message)))
  (define service-name (bit-string->bytes (ssh-msg-userauth-request-service-name message)))
  (cond
   [(and (positive? (bytes-length user-name))
	 (equal? service-name #"ssh-connection"))
    ;; TODO: Actually implement client authentication
    (prefix-transition
	(start-connection-service
	 (set-handlers (struct-copy connection conn
				    [authentication-state (authenticated user-name service-name)])
		       SSH_MSG_USERAUTH_REQUEST
		       (lambda (packet message conn)
			 ;; RFC4252 section 5.1 page 6
			 conn)))
      (send-message (outbound-packet (ssh-msg-userauth-success))))]
   [else
    (transition conn
      (send-message (outbound-packet (ssh-msg-userauth-failure '(none) #f))))]))

;; SKETCH-O
(define (start-connection-service conn)
  conn)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Session main process
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (connection-username conn)
  (match (connection-authentication-state conn)
    ((authenticated username servicename)
     username)
    (else (disconnect-with-error SSH_DISCONNECT_PROTOCOL_ERROR
				 "Not authenticated"))))

(define (bump-total t amount)
  (extend-transition* t
    (lambda (conn)
      (struct-copy connection conn
		   [total-transferred (+ (connection-total-transferred conn) amount)]))))

;; (K V A -> A) A Hash<K,V> -> A
(define (hash-fold fn seed hash)
  (do ((pos (hash-iterate-first hash) (hash-iterate-next hash pos))
       (seed seed (fn (hash-iterate-key hash pos) (hash-iterate-value hash pos) seed)))
      ((not pos) seed)))

(define (rekey-wrap t)
  (extend-transition* t
    (lambda (conn)
      (define rekey (connection-rekey-state conn))
      (if (time-to-rekey? rekey conn)
	  (let ((algs ((local-algorithm-list))))
	    (transition (struct-copy connection conn [rekey-state (rekey-local algs)])
	      (send-message (outbound-packet algs))))
	  conn))))

;; PacketDispatcher. Handles the core transport message types.
(define base-packet-dispatcher
  (hasheq SSH_MSG_DISCONNECT handle-msg-disconnect
	  SSH_MSG_IGNORE handle-msg-ignore
	  SSH_MSG_UNIMPLEMENTED handle-msg-unimplemented
	  SSH_MSG_DEBUG handle-msg-debug
	  SSH_MSG_KEXINIT handle-msg-kexinit))

(define (ssh-session local-identification-string
		     peer-identification-string
		     session-role)
  (lambda (self-pid)
    (transition (connection #f
			    base-packet-dispatcher
			    0
			    (rekey-in-seconds-or-bytes -1 -1 0)
			    #f
			    (hash)
			    (case session-role ((client) #f) ((server) #t))
			    local-identification-string
			    peer-identification-string
			    #f)
      (role 'rekey-waiter (topic-subscriber (timer-expired 'rekey-timer (wild)))
	#:state conn
	[(timer-expired 'rekey-timer now)
	 (rekey-wrap conn)])

      (role 'credit-listener (topic-subscriber (outbound-byte-credit (wild)))
	#:state conn
	[(outbound-byte-credit amount)
	 (rekey-wrap (bump-total conn amount))])

      (role 'packet-listener (topic-subscriber (inbound-packet (wild) (wild) (wild) (wild)))
	#:state conn
	[(inbound-packet sequence-number payload message transfer-size)
	 (let* ((t (if (connection-discard-next-packet? conn)
		       (struct-copy connection conn [discard-next-packet? #f])
		       (dispatch-packet sequence-number payload message conn)))
		(t (bump-total t transfer-size))
		(t (extend-transition t (send-message (inbound-credit 1)))))
	   (rekey-wrap t))]))))
